import numpy as np
from abc import ABC, abstractmethod
import optuna
import matplotlib.pyplot as plt

rng = np.random.default_rng()


class Strategy(ABC):

    @abstractmethod
    def vector(self):
        pass


class ConstantThresholdStrategy(Strategy):

    def __init__(self, N, threshold):
        self.N = N
        self.threshold = threshold
        self._vector = np.empty(2 * self.N, dtype=int)
        self._vector[:-1] = self.threshold
        self._vector[-1] = 0

    def vector(self):
        return self._vector


class LinearThresholdStrategy(Strategy):

    def __init__(self, N, a, b):
        self.N = N
        self.a = a
        self.b = b
        self._vector = np.empty(2 * self.N, dtype=int)
        self._vector = np.floor(self.a * (2 * self.N - np.arange(2 * self.N)) + self.b)
        self._vector[-1] = 0

    def vector(self):
        return self._vector


class BellmanStrategy(Strategy):

    def __init__(self, N):
        self.N = N
        self._vector = np.empty(2 * self.N, dtype=int)
        self.stop = np.full((self.N + 1, self.N + 1), False)
        self.v = np.zeros((self.N + 1, self.N + 1))
        self._compute()
        self._vector[-1] = 0

    def _compute(self):
        self.v[:, 0] = np.arange(self.N + 1)
        self.stop[0, :] = True
        for r in range(1, self.N + 1):
            for b in range(1, self.N + 1):
                continuation_value = (r / (r + b) * (1.0 + self.v[r - 1, b]) +
                                      b / (r + b) * (-1.0 + self.v[r, b - 1]))
                if continuation_value <= 0.:
                    self.stop[r, b] = True
                    self.v[r, b] = 0.
                else:
                    self.v[r, b] = continuation_value

        for n in range(1, 2*self.N):
            indices = [(i, 2*self.N - n - i) for i in range(max(0, 2*self.N- n - N), min(N, 2*self.N- n) + 1) if 0 <= 2*self.N - n - i <= N]
            rows = [i for i,j in indices]
            cols = [j for i,j in indices]
            stop_or_not = self.stop[rows, cols]

            if np.all(~stop_or_not):
                self._vector[n-1] = self.N+1
            else:
                first_false = np.argmax(~stop_or_not)
                self._vector[n-1] = cols[first_false-1] - rows[first_false-1]



    def vector(self):
        return self._vector


class GameSimulator:
    """
    docstring for GameSimulator
    """

    def __init__(self, N):
        self.N = N
        self.deck = np.array([-1] * self.N + [1] * self.N, dtype=np.int8)
        # self.deck = np.concatenate([-np.ones(self.N), np.ones(self.N)])

    def play(self, strategy):
        rng.shuffle(self.deck)
        scores = np.cumsum(self.deck)
        return scores[scores >= strategy.vector()][0]

    def monte_carlo(self, strategy, M=50000):
        mat = np.tile(self.deck, (M, 1))
        np.apply_along_axis(rng.shuffle, 1, mat)
        scores = np.cumsum(mat, axis=1)
        mask = scores >= strategy.vector()
        indices = np.argmax(mask, axis=1)
        return np.mean(scores[np.arange(M), indices])


if __name__ == "__main__":
    N = 200

    game = GameSimulator(N)

    for threshold in range(1, 8):
        s = ConstantThresholdStrategy(N, threshold)
        print(f"threshold = {threshold} --  expected score = {game.monte_carlo(s)}")

    s = LinearThresholdStrategy(N, 0.1, 2)
    print(f"(a,b) = {(s.a,s.b)} --  expected score = {game.monte_carlo(s)}")

    print(s.vector())

    def objective(trial):
        a = trial.suggest_uniform("a", 0.0, 2.0)  # x is between 0 and pi
        b = trial.suggest_uniform("b", 0.0, 5.0)  # y is between 0 and pi
        s = LinearThresholdStrategy(N, a, b)

        return -game.monte_carlo(s)

    study = optuna.create_study(direction="minimize")
    study.optimize(objective, n_trials=100)

    plt.plot(LinearThresholdStrategy(N, **study.best_params).vector())
    plt.show()

    s = LinearThresholdStrategy(N, **study.best_params)
    print(f"score obtained by playing the best optuna linear strategy = {game.monte_carlo(s)}")



    belstrat = BellmanStrategy(N)
    print(belstrat.v[-1,-1])
    plt.imshow(belstrat.stop)
    plt.show()
    plt.plot(belstrat.vector())
    plt.show()

    print(f"score obtained by playing the supposedly optimal strategy = {game.monte_carlo(belstrat)}")

